// shanChen.m: Multi-component fluid, using a LB method,
//   based on the Shan-Chen model
// [X.Shan and H.Chen, http://dx.doi.org/10.1103/PhysRevE.47.1815].
//

// Lattice Boltzmann sample, written in Matlab
// Copyright (C) 2008 Orestis Malaspinas, Andrea Parmigiani, Jonas Latt
// Address: EPFL-STI-LIN Station 9
// E-mail: orestis.malaspinas@epfl.ch

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
// Boston, MA  02110-1301, USA.

// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2014 - Scilab Enterprises - Cedric DELAMARRE
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

abs_path = get_absolute_file_path("shanchen_cpu.sce");
exec(abs_path + "cshift.sci");

txt = ["Max Iteration"; "Width"; "Height"; "Refresh Cycle"];
//userparams = x_mdialog("Shanchen demo parameters",txt,["1000";"800";"600";"1"]);
userparams = [1000;800;600;1];
if userparams == [] then
    return
end

h = scf();
colorsize = 2^14; // 16k color
h.color_map = jetcolormap(colorsize);

refresh = evstr(userparams(4));

// GENERAL FLOW CONSTANTS
ly = evstr(userparams(3));
lx = evstr(userparams(2));

G = -1.2;  // Amplitude of the molecular interaction force

omega1 = 1;  // Relaxation parameter for fluid 1
omega2 = 1;  // Relaxation parameter for fluid 2

maxT = evstr(userparams(1));    // total number of iterations

tic();

// D2Q9 LATTICE CONSTANTS
tNS   = [4/9, 1/9,1/9,1/9,1/9, 1/36,1/36,1/36,1/36];
cxNS  = [  0,   1,  0, -1,  0,    1,  -1,  -1,   1];
cyNS  = [  0,   0,  1,  0, -1,    1,   1,  -1,  -1];

drho = 0.001;
delta_rho = -drho*(1-2.0*rand(1, ly * lx));

fOut = zeros(9, ly * lx);
gOut = zeros(9, ly * lx);
fIn  = zeros(9, ly * lx);
gIn  = zeros(9, ly * lx);

// INITIAL CONDITION FOR BOTH DISTRIBUTION FUNCTIONS: (T=0) ==> TIn(i) = t(i)
for i=1:9
    fIn(i,:) = tNS(i).*(1.0 + delta_rho);
    gIn(i,:) = tNS(i).*(1.0 - delta_rho);
end

rho1 = matrix(sum(fIn, "m"),ly,lx);
img1 = abs(colorsize*rho1/max(rho1));
Matplot(img1);
e=gce();

// MAIN LOOP (TIME CYCLES)
Gomega1 = G/omega1;
Gomega2 = G/omega2;
fEq = zeros(9, ly * lx);
gEq = zeros(9, ly * lx);

for cycle = 1:maxT
    // compute rho1
    rho1 = sum(fIn, "m");
    // compute rho2
    rho2 = sum(gIn, "m");

    // compute d_rhoContrib
    rhoContrib1x = zeros(ly, lx);
    rhoContrib2x = zeros(ly, lx);
    rhoContrib1y = zeros(ly, lx);
    rhoContrib2y = zeros(ly, lx);
    for i=2:9
        work1 = rho1 * tNS(i);
        work2 = rho2 * tNS(i);
        cmw1 = cshift(matrix(work1, ly, lx), [cxNS(i),cyNS(i)]);
        cmw2 = cshift(matrix(work2, ly, lx), [cxNS(i),cyNS(i)]);
        rhoContrib1x = rhoContrib1x + cmw1 * cxNS(i);
        rhoContrib1y = rhoContrib1y + cmw1 * cyNS(i);
        rhoContrib2x = rhoContrib2x + cmw2 * cxNS(i);
        rhoContrib2y = rhoContrib2y + cmw2 * cyNS(i);
    end

    // compute d_fOut, d_gOut
    jx1 = cxNS * fIn;
    jy1 = cyNS * fIn;
    jx2 = cxNS * gIn;
    jy2 = cyNS * gIn;

    rhoTot_OMEGA = rho1 * omega1 + rho2 * omega2;
    uTotX = (jx1 * omega1 + jx2 * omega2) ./ rhoTot_OMEGA;
    uTotY = (jy1 * omega1 + jy2 * omega2) ./ rhoTot_OMEGA;

    uTotX1 = uTotX - Gomega1 .* matrix(rhoContrib2x, 1, ly * lx); //POTENTIAL CONTRIBUTION OF FLUID 2 ON 1
    uTotY1 = uTotY - Gomega1 .* matrix(rhoContrib2y, 1, ly * lx);
    uTotX2 = uTotX - Gomega2 .* matrix(rhoContrib1x, 1, ly * lx); //POTENTIAL CONTRIBUTION OF FLUID 2 ON 1
    uTotY2 = uTotY - Gomega2 .* matrix(rhoContrib1y, 1, ly * lx);

    // COLLISION STEP FLUID 1 AND 2
    for i=1:9
        cuNS1 = 3 * (cxNS(i) * uTotX1 + cyNS(i) * uTotY1);
        cuNS2 = 3 * (cxNS(i) * uTotX2 + cyNS(i) * uTotY2);
        fEq(i,:) = rho1 .* tNS(i) .* ( 1 + cuNS1 + 0.5*(cuNS1.*cuNS1) - 1.5*(uTotX1.*uTotX1+uTotY1.*uTotY1) );
        gEq(i,:) = rho2 .* tNS(i) .* ( 1 + cuNS2 + 0.5*(cuNS2.*cuNS2) - 1.5*(uTotX2.*uTotX2+uTotY2.*uTotY2) );
    end

    fOut = fIn - omega1 .* (fIn-fEq);
    gOut = gIn - omega2 .* (gIn-gEq);

    // STREAMING STEP FLUID 1 AND 2
    // compute d_fIn, d_gIn
    for i=1:9
        fIn(i,:) = matrix(cshift(matrix(fOut(i,:), ly, lx), [cxNS(i),cyNS(i)]), 1, ly * lx);
        gIn(i,:) = matrix(cshift(matrix(gOut(i,:), ly, lx), [cxNS(i),cyNS(i)]), 1, ly * lx);
    end

    // VISUALIZATION
    if modulo(cycle, refresh) == 0 then
        h.info_message = "Fluid 1 density : HOST mode : Cycle ="+string(cycle)+"/"+string(maxT)+" Elapsed time : "+string(toc())+"s";
        rho1 = matrix(rho1,ly,lx);
        img1 = abs(colorsize*rho1/max(rho1));
        if is_handle_valid(e) == %f then
            return;
        end
        e.data = img1;
    end
end

