// shanChen.m: Multi-component fluid, using a LB method,
//   based on the Shan-Chen model
// [X.Shan and H.Chen, http://dx.doi.org/10.1103/PhysRevE.47.1815].
//

// Lattice Boltzmann sample, written in Matlab
// Copyright (C) 2008 Orestis Malaspinas, Andrea Parmigiani, Jonas Latt
// Address: EPFL-STI-LIN Station 9
// E-mail: orestis.malaspinas@epfl.ch

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
// Boston, MA  02110-1301, USA.

// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2014 - Scilab Enterprises - Cedric DELAMARRE
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

abs_path = get_absolute_file_path("shanchen_gpu.sce");

txt = ["Max Iteration"; "Width"; "Height"; "Platform used"; "Device used"; "Refresh Cycle"];
userparams = x_mdialog("Shanchen demo parameters",txt,["1000";"800";"600";"0";"0";"1"]);

if userparams == [] then
    return
end

gpuExit();
gpuInit(evstr(userparams(5)), evstr(userparams(4)));
[device platform name] = gpuUsed();
disp("demo running on "+name);
refresh = evstr(userparams(6));

// GENERAL FLOW CONSTANTS
// max size for work item : nbMaxThread x nbMaxThread per work group
nbMaxThread = 16;
lw = evstr(userparams(2));
lh = evstr(userparams(3));

h = scf();
colorsize = 256;//2^14; // 16k color
h.color_map = jetcolormap(colorsize);

G = -1.2;  // Amplitude of the molecular interaction force

omega1 = 1;  // Relaxation parameter for fluid 1
omega2 = 1;  // Relaxation parameter for fluid 2

maxT   = evstr(userparams(1));    // total number of iterations

tic();

// D2Q9 LATTICE CONSTANTS
tNS   = [4/9, 1/9,1/9,1/9,1/9, 1/36,1/36,1/36,1/36];
cxNS  = [  0,   1,  0, -1,  0,    1,  -1,  -1,   1];
cyNS  = [  0,   0,  1,  0, -1,    1,   1,  -1,  -1];

drho = 0.001;
delta_rho = -drho*(1-2.0*rand(1, lh * lw));

// INITIAL CONDITION FOR BOTH DISTRIBUTION FUNCTIONS: (T=0) ==> TIn(i) = t(i)
fIn  = zeros(9, lh * lw);
gIn  = zeros(9, lh * lw);
for i=1:9
    fIn(i,:) = tNS(i).*(1.0 + delta_rho);
    gIn(i,:) = tNS(i).*(1.0 - delta_rho);
end

rho1 = matrix(sum(fIn, "m"),lh,lw);
img1 = abs(colorsize*rho1/max(rho1));
Matplot(img1);
e=gce();

if lw < nbMaxThread then
    nb_threadX = lw;
else
    nb_threadX = nbMaxThread;
end

if lh < nbMaxThread then
    nb_threadY = lh;
else
    nb_threadY = nbMaxThread;
end

nb_blockY = ceil(lh / nb_threadX);
nb_blockX = ceil(lw / nb_threadY);

bin             = gpuBuild(abs_path + "shanchen_kernels");
cpt_sum         = gpuLoadFunction(bin, "compute_sum", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
cpt_work        = gpuLoadFunction(bin, "compute_work", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
cpt_rhoContrib  = gpuLoadFunction(bin, "compute_rhoContrib", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
cpt_Out         = gpuLoadFunction(bin, "compute_Out", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
cpt_In          = gpuLoadFunction(bin, "compute_In", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
disp("Kernels build and load : "+string(toc())+" sec")

d_fIn   = gpuSetData(fIn);
d_gIn   = gpuSetData(gIn);
d_rho1  = gpuAlloc(lh, lw);
d_rho2  = gpuAlloc(lh, lw);
d_work1 = gpuAlloc(9, lh * lw);
d_work2 = gpuAlloc(9, lh * lw);
d_fOut  = gpuAlloc(9, lh * lw);
d_gOut  = gpuAlloc(9, lh * lw);
d_cxNS  = gpuSetData(cxNS);
d_cyNS  = gpuSetData(cyNS);
d_tNS   = gpuSetData(tNS);

d_rhoContrib = gpuAlloc(4, lh * lw);

// MAIN LOOP (TIME CYCLES)
for cycle = 1:maxT
    // compute rho1
    cpt_sum(d_fIn, d_rho1, int32(lh), int32(lw));

    // compute rho2
    cpt_sum(d_gIn, d_rho2, int32(lh), int32(lw));

    // compute d_work1, d_work2,
    cpt_work(d_rho1, d_rho2, d_cxNS, d_cyNS, d_tNS, d_work1, d_work2, int32(lh), int32(lw));

    // compute d_rhoContrib
    cpt_rhoContrib(d_work1, d_work2, d_cxNS, d_cyNS, d_rhoContrib, int32(lh), int32(lw));

    // compute d_fOut, d_gOut
    cpt_Out(d_fIn, d_gIn, d_rho1, d_rho2, d_cxNS, d_cyNS, d_tNS, d_rhoContrib, ..
               d_fOut, d_gOut, G, omega1, omega2, int32(lh), int32(lw));

    // compute d_fIn, d_gIn
    cpt_In(d_fIn, d_gIn, d_cxNS, d_cyNS, d_fOut, d_gOut, int32(lh), int32(lw));

    // VISUALIZATION
    if modulo(cycle, refresh) == 0 then
        rho1 = gpuGetData(d_rho1);
        img1 = abs(colorsize*rho1/max(rho1));
        if is_handle_valid(e) then
            e.data = img1;
            h.info_message = "Fluid 1 density : "+ gpuGetMode() +" mode : Cycle ="+string(cycle)+"/"+string(maxT)+" Elapsed time : "+string(toc())+"s ";
        else
            return;
        end
    end
end

disp("Total execution time : "+string(toc())+" sec");

clear d_fIn d_gIn d_rho1 d_rho2 d_work1 d_work2 d_fOut d_gOut d_cxNS d_cyNS d_tNS d_rhoContrib
