/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) 2014 - Scilab Enterprises - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

extern "C"
{
__global__ void compute_sum(double* fIn, double* rho, int iRows, int iCols)
{
    int idCols = threadIdx.x + blockIdx.x * blockDim.x;
    int idRows = threadIdx.y + blockIdx.y * blockDim.y;

    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int id = idCols * iRows + idRows;
    int iPos = id * 9;

    rho[id] = fIn[iPos];
    for (int i = 1; i < 9; i++)
    {
        rho[id] += fIn[iPos + i];
    }
}

__global__ void compute_work(double* rho1, double* rho2,
                             double* cxNS, double* cyNS,
                             double* tNS,
                             double* work1, double* work2,
                             int iRows, int iCols)
{
    int idCols = threadIdx.x + blockIdx.x * blockDim.x;
    int idRows = threadIdx.y + blockIdx.y * blockDim.y;
    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int cshiftPos = 0;
    int id = idCols * iRows + idRows;
    for (int i = 0; i < 9; i++)
    {
        double val1 = rho1[id] * tNS[i];
        double val2 = rho2[id] * tNS[i];

        // circshift management
        int iPosRows = idRows + cxNS[i];
        int iPosCols = idCols + cyNS[i];

        if(iPosRows >= iRows)
        {
            iPosRows -= iRows;
        }

        if(iPosRows < 0)
        {
            iPosRows += iRows;
        }

        if(iPosCols >= iCols)
        {
            iPosCols -= iCols;
        }

        if(iPosCols < 0)
        {
            iPosCols += iCols;
        }

        cshiftPos = iPosCols * iRows + iPosRows;
        work1[cshiftPos * 9 + i] = val1;
        work2[cshiftPos * 9 + i] = val2;
    }
}

__global__ void compute_rhoContrib(double* work1, double* work2,
                                   double* cxNS, double* cyNS,
                                   double* rhoContrib,
                                   int iRows, int iCols)
{
    int idCols = threadIdx.x + blockIdx.x * blockDim.x;
    int idRows = threadIdx.y + blockIdx.y * blockDim.y;
    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int id = idCols * iRows + idRows;
    int iPosW = id * 9;
    int iPosR = id * 4;

    rhoContrib[iPosR]     = 0;
    rhoContrib[iPosR + 1] = 0;
    rhoContrib[iPosR + 2] = 0;
    rhoContrib[iPosR + 3] = 0;

    for (int i = 1; i < 9; i++)
    {
        rhoContrib[iPosR]     += work1[iPosW + i] * cxNS[i];
        rhoContrib[iPosR + 1] += work1[iPosW + i] * cyNS[i];
        rhoContrib[iPosR + 2] += work2[iPosW + i] * cxNS[i];
        rhoContrib[iPosR + 3] += work2[iPosW + i] * cyNS[i];
    }
}

__global__ void compute_Out(double* fIn, double* gIn,
                            double* rho1, double* rho2,
                            double* cxNS, double* cyNS,
                            double* tNS, double* rhoContrib,
                            double* fOut, double* gOut,
                            double G, double omega1, double omega2, int iRows, int iCols)
{
    int idCols = threadIdx.x + blockIdx.x * blockDim.x;
    int idRows = threadIdx.y + blockIdx.y * blockDim.y;
    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    double jx1      = 0;
    double jy1      = 0;
    double jx2      = 0;
    double jy2      = 0;
    double fEq      = 0;
    double gEq      = 0;
    double omega    = 0;
    double uTotX    = 0;
    double uTotY    = 0;
    double uTotX1   = 0;
    double uTotY1   = 0;
    double uTotX2   = 0;
    double uTotY2   = 0;
    double cuNS1    = 0;
    double cuNS2    = 0;

    int id = idCols * iRows + idRows;
    int iPos = id * 9;
    for (int i = 0; i < 9; i++)
    {
        jx1 += cxNS[i] * fIn[iPos + i];
        jy1 += cyNS[i] * fIn[iPos + i];
        jx2 += cxNS[i] * gIn[iPos + i];
        jy2 += cyNS[i] * gIn[iPos + i];
    }

    omega = rho1[id] * omega1 + rho2[id] * omega2;
    uTotX = (jx1 * omega1 + jx2 * omega2) / omega;
    uTotY = (jy1 * omega1 + jy2 * omega2) / omega;

    uTotX1 = uTotX - (G / omega1) * rhoContrib[id * 4 + 2];
    uTotY1 = uTotY - (G / omega1) * rhoContrib[id * 4 + 3];
    uTotX2 = uTotX - (G / omega2) * rhoContrib[id * 4];
    uTotY2 = uTotY - (G / omega2) * rhoContrib[id * 4 + 1];

    for (int i = 0; i < 9; i++)
    {
        cuNS1 = 3 * (cxNS[i] * uTotX1 + cyNS[i] * uTotY1);
        cuNS2 = 3 * (cxNS[i] * uTotX2 + cyNS[i] * uTotY2);
        fEq = rho1[id] * tNS[i] * ( 1 + cuNS1 + 0.5 * (cuNS1 * cuNS1) - 1.5 * (uTotX1 * uTotX1 + uTotY1 * uTotY1) );
        gEq = rho2[id] * tNS[i] * ( 1 + cuNS2 + 0.5 * (cuNS2 * cuNS2) - 1.5 * (uTotX2 * uTotX2 + uTotY2 * uTotY2) );
        fOut[iPos + i] = fIn[iPos + i] - omega1 * (fIn[iPos + i] - fEq);
        gOut[iPos + i] = gIn[iPos + i] - omega2 * (gIn[iPos + i] - gEq);
    }
}

__global__ void compute_In(double* fIn, double* gIn,
                           double* cxNS, double* cyNS,
                           double* fOut, double* gOut,
                           int iRows, int iCols)
{
    int idCols = threadIdx.x + blockIdx.x * blockDim.x;
    int idRows = threadIdx.y + blockIdx.y * blockDim.y;
    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int cshiftPos = 0;
    int iPos = (idCols * iRows + idRows) * 9;
    for (int i = 0; i < 9; i++)
    {
        // circshift management
        int iPosRows = idRows + cxNS[i];
        int iPosCols = idCols + cyNS[i];

        if(iPosRows >= iRows)
        {
            iPosRows -= iRows;
        }

        if(iPosRows < 0)
        {
            iPosRows += iRows;
        }

        if(iPosCols >= iCols)
        {
            iPosCols -= iCols;
        }

        if(iPosCols < 0)
        {
            iPosCols += iCols;
        }

        cshiftPos = iPosCols * iRows + iPosRows;
        fIn[cshiftPos * 9 + i] = fOut[iPos + i];
        gIn[cshiftPos * 9 + i] = gOut[iPos + i];
    }
}
}
