// shanChen.m: Multi-component fluid, using a LB method,
//   based on the Shan-Chen model
// [X.Shan and H.Chen, http://dx.doi.org/10.1103/PhysRevE.47.1815].
//

// Lattice Boltzmann sample, written in Matlab
// Copyright (C) 2008 Orestis Malaspinas, Andrea Parmigiani, Jonas Latt
// Address: EPFL-STI-LIN Station 9
// E-mail: orestis.malaspinas@epfl.ch

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
// Boston, MA  02110-1301, USA.

// Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
// Copyright (C) 2014 - Scilab Enterprises - Cedric DELAMARRE
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

// mpi command line
// mpirun -n 3 "ScilabInstallPath\scilab" -f "sciGPGPUInstallPath\demos\shanchen\shanchen_mpi.sce"
// OR
// mpirun -n 1 "ScilabInstallPath\bin\scilab" -noatomsautoload -f "sciGPGPUInstallPath\demos\shanchen\shanchen_mpi.sce" : -n 2 "ScilabInstallPath\bin\scilab" -nwni -f "sciGPGPUInstallPath\demos\shanchen\shanchen_mpi.sce"

abs_path = get_absolute_file_path("shanchen_mpi.sce");

MPI_Init();
rnk     = MPI_Comm_rank();
nbSlave = MPI_Comm_size() - 1;
Slave   = rnk;
Master  = ~rnk;

marginWidth = 2;

if Master then
    txt = ["Max Iteration"; "Width"; "Height"; "Refresh Cycle"];
    userparams = x_mdialog("Shanchen demo parameters",txt,["1000";"800";"600";"20"]);

    if userparams == [] then
        return
    end

    h = scf();
    colorsize = 2^14; // 16k color
    h.color_map = jetcolormap(colorsize);

    refresh = evstr(userparams(4));
    lw = evstr(userparams(2));
    lh = evstr(userparams(3));
    maxT = evstr(userparams(1));    // total number of iterations
    tic();

    lw2 = ceil(lw / nbSlave);
    for slaveId = 1:nbSlave
        // send Width
        MPI_Send(lw2 + 2 * marginWidth, slaveId);
        // send Height
        MPI_Send(lh, slaveId);
        // send maxT
        MPI_Send(maxT, slaveId);
        // send refresh
        MPI_Send(refresh, slaveId);
    end

    tNS         = [4/9, 1/9,1/9,1/9,1/9, 1/36,1/36,1/36,1/36];
    drho        = 0.001;
    delta_rho   = -drho*(1-2.0*rand(1, lw * lh));

    // INITIAL CONDITION FOR BOTH DISTRIBUTION FUNCTIONS: (T=0) ==> TIn(i) = t(i)
    fIn = zeros(9, lw * lh);
    gIn = zeros(9, lw * lh);
    for i=1:9
        fIn(i,:) = tNS(i).*(1.0 + delta_rho);
        gIn(i,:) = tNS(i).*(1.0 - delta_rho);
    end

    rho1 = matrix(sum(fIn, "m"), lh, lw);
    img1 = abs(colorsize*rho1/max(rho1));
    Matplot(img1);
    e=gce();

    XSizePart  = lw2;
    sizePart   = 9 * lh * XSizePart;
    sizeMargin = 9 * lh * marginWidth;

    // compute position of margin
    marginBegin = zeros(nbSlave, 2);
    marginEnd   = zeros(nbSlave, 2);
    mb = 0;
    me = sizeMargin + sizePart;
    for i = 1:nbSlave
        marginBegin(i, 1) = mb + 1;
        marginBegin(i, 2) = mb + sizeMargin;
        marginEnd(i, 1)   = me + 1;
        marginEnd(i, 2)   = me + sizeMargin;
        mb = me - sizeMargin;
        me = me + sizePart;
    end

    fInWork = zeros(9, (lw + 2 * marginWidth) * lh);
    gInWork = zeros(9, (lw + 2 * marginWidth) * lh);

    // copy fIn in the middle of fInWork
    fInWork((sizeMargin + 1):($ - sizeMargin)) = fIn(:);
    // copy fIn in the middle of fInWork
    gInWork((sizeMargin + 1):($ - sizeMargin)) = gIn(:);

    // SEND DATA
    for slaveId = 1:nbSlave
        MPI_Send(fInWork(marginBegin(slaveId, 1) : marginEnd(slaveId, 2)) ,slaveId);
        MPI_Send(gInWork(marginBegin(slaveId, 1) : marginEnd(slaveId, 2)) ,slaveId);
    end

    for cycle = 1:maxT
        // rewrote last colon in first margin and first colon in last margin
        fInWork(1:sizeMargin)       = fInWork(marginEnd($,1)-sizeMargin:marginEnd($,1)-1);
        fInWork($-sizeMargin+1:$)   = fInWork(marginBegin(1,2)+1:marginBegin(1,2)+sizeMargin);
        gInWork(1:sizeMargin)       = gInWork(marginEnd($,1)-sizeMargin:marginEnd($,1)-1);
        gInWork($-sizeMargin+1:$)   = gInWork(marginBegin(1,2)+1:marginBegin(1,2)+sizeMargin);

        // SEND DATA
        for slaveId = 1:nbSlave
            MPI_Send(fInWork(marginBegin(slaveId, 1) : marginBegin(slaveId, 2)) ,slaveId);
            MPI_Send(fInWork(marginEnd(slaveId, 1)   : marginEnd(slaveId, 2)) ,slaveId);
            MPI_Send(gInWork(marginBegin(slaveId, 1) : marginBegin(slaveId, 2)) ,slaveId);
            MPI_Send(gInWork(marginEnd(slaveId, 1)   : marginEnd(slaveId, 2)) ,slaveId);
        end

        // RECEIVE DATA
        for slaveId = 1:nbSlave
            m1 = marginBegin(slaveId, 2)+1 : marginBegin(slaveId, 2) + sizeMargin;
            m2 = marginEnd(slaveId, 1)-sizeMargin : marginEnd(slaveId, 1)-1;
            fInWork(m1) = MPI_Recv(slaveId, 0);
            fInWork(m2) = MPI_Recv(slaveId, 0);
            gInWork(m1) = MPI_Recv(slaveId, 0);
            gInWork(m2) = MPI_Recv(slaveId, 0);
        end

        // Display result
        if modulo(cycle, refresh) == 0 then
            posBeg = 1;
            posEnd = lh * XSizePart;
            for slaveId = 1:nbSlave
                rho1(posBeg:posEnd) = MPI_Recv(slaveId, 0);
                posBeg = posBeg + lh * XSizePart;
                posEnd = posEnd + lh * XSizePart;
            end

            img1 = abs(colorsize*rho1/max(rho1));
            if is_handle_valid(e) then
                e.data = img1;
                h.info_message = "Fluid 1 density : MPI + GPU : Cycle ="+string(cycle)+"/"+string(maxT)+" Elapsed time : "+string(toc())+"s ";
                for slaveId = 1:nbSlave
                    MPI_Send(0 ,slaveId);
                end
            else
                for slaveId = 1:nbSlave
                    MPI_Send(1 ,slaveId);
                end
                MPI_Finalize();
                exit();
            end
        end
    end

    for slaveId = 1:nbSlave
        MPI_Send(1 ,slaveId);
    end

    MPI_Finalize();
end

if Slave then
    Master = 0;

//    exec(abs_path + "../../loader.sce");
    gpuExit();
    idDevice = gpuInit(Slave-1);

    disp("Slave "+string(Slave)+" use Device "+string(idDevice));

    // max size for work item : nbMaxThread x nbMaxThread per work group
    nbMaxThread = 16;

    lw      = MPI_Recv(Master, 0);
    lh      = MPI_Recv(Master, 0);
    maxT    = MPI_Recv(Master, 0);
    refresh = MPI_Recv(Master, 0);

    G = -1.2;  // Amplitude of the molecular interaction force

    omega1 = 1;  // Relaxation parameter for fluid 1
    omega2 = 1;  // Relaxation parameter for fluid 2

    // D2Q9 LATTICE CONSTANTS
    tNS   = [4/9, 1/9,1/9,1/9,1/9, 1/36,1/36,1/36,1/36];
    cxNS  = [  0,   1,  0, -1,  0,    1,  -1,  -1,   1];
    cyNS  = [  0,   0,  1,  0, -1,    1,   1,  -1,  -1];

    if lw < nbMaxThread then
        nb_threadX = lw;
    else
        nb_threadX = nbMaxThread;
    end

    if lh < nbMaxThread then
        nb_threadY = lh;
    else
        nb_threadY = nbMaxThread;
    end

    nb_blockY = ceil(lh / nb_threadX);
    nb_blockX = ceil(lw / nb_threadY);

    bin             = gpuBuild(abs_path + "shanchen_kernels", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
    cpt_sum         = gpuLoadFunction(bin, "compute_sum", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
    cpt_work        = gpuLoadFunction(bin, "compute_work", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
    cpt_rhoContrib  = gpuLoadFunction(bin, "compute_rhoContrib", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
    cpt_Out         = gpuLoadFunction(bin, "compute_Out", nb_threadY, nb_threadX, nb_blockY, nb_blockX);
    cpt_In          = gpuLoadFunction(bin, "compute_In", nb_threadY, nb_threadX, nb_blockY, nb_blockX);

    d_rho1  = gpuAlloc(lh, lw);
    d_rho2  = gpuAlloc(lh, lw);
    d_work1 = gpuAlloc(9, lh * lw);
    d_work2 = gpuAlloc(9, lh * lw);
    d_fOut  = gpuAlloc(9, lh * lw);
    d_gOut  = gpuAlloc(9, lh * lw);
    d_cxNS  = gpuSetData(cxNS);
    d_cyNS  = gpuSetData(cyNS);
    d_tNS   = gpuSetData(tNS);

    d_rhoContrib = gpuAlloc(4, lh * lw);

    sizeMargin = 9 * lh * marginWidth;

    fIn = MPI_Recv(Master, 0);
    gIn = MPI_Recv(Master, 0);

    d_fIn = gpuSetData(fIn);
    d_gIn = gpuSetData(gIn);

    iLastPos = size(fIn, "*");
    RecvPosB = gpuSetData(1:sizeMargin);
    RecvPosE = gpuSetData(iLastPos-sizeMargin+1:iLastPos);
    SendPosB = gpuSetData(sizeMargin+1:2*sizeMargin);
    SendPosE = gpuSetData(iLastPos-sizeMargin*2+1:iLastPos-sizeMargin);

    rho1Size = lh * lw;
    rho1sizeMargin = lh * marginWidth;
    rho1DataPos = gpuSetData(rho1sizeMargin+1:rho1Size-rho1sizeMargin);

    // MAIN LOOP (TIME CYCLES)
    goOut = 0;
    cycle = 0;
    while goOut == 0 & cycle <> maxT
        cycle = cycle + 1;

        d_fIn(RecvPosB) = MPI_Recv(Master, 0);
        d_fIn(RecvPosE) = MPI_Recv(Master, 0);
        d_gIn(RecvPosB) = MPI_Recv(Master, 0);
        d_gIn(RecvPosE) = MPI_Recv(Master, 0);

        // compute rho1
        cpt_sum(d_fIn, d_rho1, int32(lh), int32(lw));

        // compute rho2
        cpt_sum(d_gIn, d_rho2, int32(lh), int32(lw));

        // compute d_work1, d_work2,
        cpt_work(d_rho1, d_rho2, d_cxNS, d_cyNS, d_tNS, d_work1, d_work2, int32(lh), int32(lw));

        // compute d_rhoContrib
        cpt_rhoContrib(d_work1, d_work2, d_cxNS, d_cyNS, d_rhoContrib, int32(lh), int32(lw));

        // compute d_fOut, d_gOut
        cpt_Out(d_fIn, d_gIn, d_rho1, d_rho2, d_cxNS, d_cyNS, d_tNS, d_rhoContrib, ..
                d_fOut, d_gOut, G, omega1, omega2, int32(lh), int32(lw));

        // compute d_fIn, d_gIn
        cpt_In(d_fIn, d_gIn, d_cxNS, d_cyNS, d_fOut, d_gOut, int32(lh), int32(lw));

        fInExtractB = d_fIn(RecvPosB);
        fInExtractE = d_fIn(RecvPosE);
        gInExtractB = d_gIn(RecvPosB);
        gInExtractE = d_gIn(RecvPosE);

        MPI_Send(gpuGetData(fInExtractB), Master);
        MPI_Send(gpuGetData(fInExtractE), Master);
        MPI_Send(gpuGetData(gInExtractB), Master);
        MPI_Send(gpuGetData(gInExtractE), Master);

        clear fInExtractB;
        clear fInExtractE;
        clear gInExtractB;
        clear gInExtractE;

        if modulo(cycle, refresh) == 0 then
            rho1Extract = d_rho1(rho1DataPos);
            MPI_Send(gpuGetData(rho1Extract), Master);
            clear rho1Extract;
            goOut = MPI_Recv(Master, 0);
        end
    end

    clear d_fIn;
    clear d_gIn;
    clear d_rho1;
    clear d_rho2;
    clear d_work1;
    clear d_work2;
    clear d_fOut;
    clear d_gOut;
    clear d_cxNS;
    clear d_cyNS;
    clear d_tNS;
    clear d_rhoContrib;
    gpuExit();
    MPI_Finalize();
    exit();
end


